@@ -17,8 +17,8 @@
       return string.replace(/&/g, "&amp;")
                    .replace(/</g, "&lt;")
                    .replace(/>/g, "&gt;")
-                   .replace(/"/, "&quot;")
-                   .replace(/'/, "&#39;")
+                   .replace(/"/g, "&quot;")
+                   .replace(/'/g, "&#39;")
                    ;
     },
 
